"""Simple energy-based voice activity detector."""

import array
import math
import statistics
from typing import List, Optional

_SAMPLE_RATE = 16000
_SAMPLE_WIDTH = 2  # 16-bit


class EnergyVad:
    """Energy-based voice activity detector (VAD)."""

    def __init__(
        self,
        threshold: Optional[float] = None,
        samples_per_chunk: int = 240,
        calibrate_seconds: float = 0.5,
    ) -> None:
        """
        Initialize VAD.

        Parameters
        ----------
        threshold: float, optional
            Energy threshold, above which is considered speech.
            If None, it will be set during calibration.
        samples_per_chunk: int
            Number of samples that process_chunk will expect.
        calibrate_seconds: float
            Seconds of audio that will be used to calibrate threshold.
        """
        self.samples_per_chunk = samples_per_chunk
        self.bytes_per_chunk = self.samples_per_chunk * _SAMPLE_WIDTH
        self.seconds_per_chunk = self.samples_per_chunk / _SAMPLE_RATE

        self.threshold = threshold

        self.calibrate_seconds = calibrate_seconds
        self._calibrate_seconds_left = self.calibrate_seconds
        self._calibrate_energies: List[float] = []

    def reset_calibration(self) -> None:
        """Reset threshold and calibration time."""
        self._calibrate_seconds_left = self.calibrate_seconds
        self._calibrate_energies.clear()
        self.threshold = None

    def process_chunk(self, chunk: bytes) -> Optional[bool]:
        """
        Process a single chunk of audio.

        Parameters
        ----------
        chunk: bytes
            Chunk of 16-bit mono PCM audio at 16Khz.
            Must be exactly self.bytes_per_chunk in size.

        Returns
        -------
        bool, optional
            None if chunk was used for calibration.
            True if energy > threshold, False otherwise.
        """
        if len(chunk) != self.bytes_per_chunk:
            raise ValueError(f"Chunk must be {self.bytes_per_chunk} bytes")

        chunk_array = array.array("h", chunk)

        # Compute RMS
        energy = -math.sqrt(sum(x**2 for x in chunk_array) / len(chunk_array))
        debiased_energy = math.sqrt(
            sum((x + energy) ** 2 for x in chunk_array) / len(chunk_array)
        )

        if self.threshold is None:
            if self._calibrate_seconds_left <= 0:
                # Calibration complete
                self.threshold = statistics.mean(self._calibrate_energies) + math.sqrt(
                    statistics.variance(self._calibrate_energies)
                )
            else:
                self._calibrate_energies.append(debiased_energy)
                self._calibrate_seconds_left -= self.seconds_per_chunk

            # Calibrating
            return None

        return debiased_energy > self.threshold



"""Tests for energy_vad."""

import wave
from pathlib import Path
from typing import List

import pytest

from energy_vad import EnergyVad

_DIR = Path(__file__).parent


def read_wav(file_name: str) -> bytes:
    """Return audio bytes from WAV file (16Khz, 16-bit mono)."""
    with wave.open(str(_DIR / file_name), "rb") as wav_file:
        assert wav_file.getframerate() == 16000
        assert wav_file.getsampwidth() == 2
        assert wav_file.getnchannels() == 1

        return wav_file.readframes(wav_file.getnframes())


def run_vad(audio: bytes) -> List[bool]:
    """Run VAD on audio and return speech/silence for each chunk."""
    vad = EnergyVad()
    results = []
    offset = 0
    while (offset + vad.bytes_per_chunk) < len(audio):
        chunk = audio[offset : offset + vad.bytes_per_chunk]
        result = vad.process_chunk(chunk)
        if result is not None:
            results.append(result)
        offset += vad.bytes_per_chunk

    return results


def test_silence() -> None:
    """Test on WAV with no speech."""
    audio = read_wav("silence.wav")
    is_speech = run_vad(audio)
    assert all(not v for v in is_speech)


def test_speech() -> None:
    """Test on WAV with some speech."""
    audio = read_wav("speech.wav")
    is_speech = run_vad(audio)
    assert any(v for v in is_speech)


def test_bad_chunk_size() -> None:
    """Test chunk size requirement in process_chunk."""
    vad = EnergyVad()
    with pytest.raises(ValueError):
        vad.process_chunk(bytes(vad.bytes_per_chunk + 1))
